/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jbpm.taskmgmt.exe;

import java.util.List;

import org.jbpm.db.AbstractDbTestCase;
import org.jbpm.graph.def.ProcessDefinition;
import org.jbpm.graph.exe.Comment;
import org.jbpm.graph.exe.ProcessInstance;


public class TaskInstanceDbTest extends AbstractDbTestCase {

  public void testTaskInstanceUnrelatedToAProcess() {
    TaskInstance taskInstance = new TaskInstance("do laundry", "someoneelse");
    session.save(taskInstance);
    long id = taskInstance.getId();
    
    newTransaction();
    
    taskInstance = (TaskInstance) session.load(TaskInstance.class, new Long(id));
    assertNotNull(taskInstance);
    assertEquals("do laundry", taskInstance.getName());
    assertEquals("someoneelse", taskInstance.getActorId());
  }

  public void testTaskInstanceBasicLifeCycle() {
    ProcessDefinition processDefinition = ProcessDefinition.parseXmlString(
      "<process-definition>" +
      "  <start-state>" +
      "    <transition to='a' />" +
      "  </start-state>" +
      "  <task-node name='a'>" +
      "    <task name='clean ceiling' />" +
      "    <transition to='end' />" +
      "  </task-node>" +
      "  <end-state name='end' />" +
      "</process-definition>"
    );

    processDefinition = saveAndReload(processDefinition);
    
    ProcessInstance processInstance = new ProcessInstance(processDefinition);
    processInstance.signal();
    
    processInstance = saveAndReload(processInstance);

    long tokenId = processInstance.getRootToken().getId();
    List taskInstances = taskMgmtSession.findTaskInstancesByToken(tokenId);
    assertEquals(1, taskInstances.size());
    TaskInstance taskInstance = (TaskInstance) taskInstances.get(0);
    assertFalse(taskInstance.hasEnded());
    assertEquals(tokenId, taskInstance.getToken().getId());
    // do some updates
    taskInstance.end();
    
    processInstance = saveAndReload(processInstance);
    
    taskInstance = taskMgmtSession.loadTaskInstance(taskInstance.getId());
    assertTrue(taskInstance.hasEnded());
  }

  public void testTaskName() {
    ProcessDefinition processDefinition = ProcessDefinition.parseXmlString(
      "<process-definition>" +
      "  <start-state>" +
      "    <transition to='a' />" +
      "  </start-state>" +
      "  <task-node name='a'>" +
      "    <task name='clean ceiling' />" +
      "    <transition to='end' />" +
      "  </task-node>" +
      "  <end-state name='end' />" +
      "</process-definition>"
    );

    processDefinition = saveAndReload(processDefinition);
    
    ProcessInstance processInstance = new ProcessInstance(processDefinition);
    processInstance.signal();
    
    processInstance = saveAndReload(processInstance);

    long tokenId = processInstance.getRootToken().getId();
    List taskInstances = taskMgmtSession.findTaskInstancesByToken(tokenId);
    assertEquals(1, taskInstances.size());
    TaskInstance taskInstance = (TaskInstance) taskInstances.get(0);
    assertFalse(taskInstance.hasEnded());
    assertEquals("clean ceiling", taskInstance.getName());
    assertEquals("clean ceiling", taskInstance.getTask().getName());
    // do some updates
    taskInstance.setName("clean ceiling thoroughly");
    
    processInstance = saveAndReload(processInstance);
    taskInstance = taskMgmtSession.loadTaskInstance(taskInstance.getId());
    
    assertEquals("clean ceiling thoroughly", taskInstance.getName());
    assertEquals("clean ceiling", taskInstance.getTask().getName());
  }
  
  public void testTaskComments() {
    ProcessDefinition processDefinition = ProcessDefinition.parseXmlString(
      "<process-definition>" +
      "  <start-state>" +
      "    <transition to='a' />" +
      "  </start-state>" +
      "  <task-node name='a'>" +
      "    <task name='clean ceiling' />" +
      "    <transition to='end' />" +
      "  </task-node>" +
      "  <end-state name='end' />" +
      "</process-definition>"
    );

    processDefinition = saveAndReload(processDefinition);
    
    ProcessInstance processInstance = new ProcessInstance(processDefinition);
    processInstance.signal();
    
    processInstance = saveAndReload(processInstance);

    long tokenId = processInstance.getRootToken().getId();
    List taskInstances = taskMgmtSession.findTaskInstancesByToken(tokenId);
    assertEquals(1, taskInstances.size());
    TaskInstance taskInstance = (TaskInstance) taskInstances.get(0);
    taskInstance.addComment("please hurry!");
    
    processInstance = saveAndReload(processInstance);
    taskInstance = taskMgmtSession.loadTaskInstance(taskInstance.getId());
    
    List comments = taskInstance.getComments();
    assertEquals(1, comments.size());
    
    Comment comment = (Comment) comments.get(0); 
    assertEquals("please hurry!", comment.getMessage());
    assertSame(taskInstance, comment.getTaskInstance());
  }
}
